#include <stdint.h>
#include <stdio.h>
#include "Pi3_serial.h"
#include "mca_serial_device.h"
#include "bpi_globals.h"


/*
  This is a simple C-driver for communicating with a BPI MCA device.
  It is intended for embedded microcontrollers with limited RAM and program space.
  
  We assume that the controller is a 32-bit device and the natural item size is 32 bit.
  
  The structure of the first 32-bit word of any command is the same for all BPI MCA devices, such as PMT-MCA, armBase/armMorpho, SiPM-MCA, and SiPM-Morpho
  
  Since the stack memory space is quite limited on a micro-controller, the generic interface assumes that all bulk data arrays are global variables.
  
  Variable sizes are shown explicitly via uint32_t, uint16_t, etc instructions.
*/

/*
    Host wants to write a command consisting of 64 x 4-byte data to the MCA.
    pSrc: Pointer to a 4-byte aligned data source
    num_words: Number of 4-bye data to be written; data are uint32_t.
    
    Data will be transmitted byte-wise with the LSB being sent first
    
    We prepare the data to be sent.  Then we send a ping to the device.
    When we receive an answer to the ping, we send the command data.
*/
void bpi_write_command(uint32_t *pSrc){
    uint8_t ping_send = 0xA5;  // A ping byte to send
    uint8_t ping_recv = 1;     // Store the received ping byte
    const uint32_t cmd_bytes = 64;  // Commands are always 64 bytes long

    bpi_serial_reset_buffers();                      // Empty buffers
    bpi_serial_write((uint8_t *)&ping_send, 1);      // Send a ping
    bpi_serial_read((uint8_t *)&ping_recv, 1);       // Blocking read of 1 byte
    bpi_serial_reset_buffers();                      // Empty buffers
    bpi_serial_write((uint8_t *)pSrc, cmd_bytes);    // Send the command data
}

/*
    Host wants to write an array of 4-byte data to the MCA, 
    after having written the command to the MCA
    pSrc: Pointer to a 4-byte aligned data source
    num_words: Number of 4-bye data to be written; data may be uint32_t or float.
    
    Data will be transmitted byte-wise with the LSB being sent first
    At this stage it is assumed that the MCA is ready to receive the data to be sent.
*/
void bpi_write_buffer(uint32_t *pSrc, uint32_t num_bytes){
    const uint32_t chunk_size = 256;
    uint8_t *pSnd = (uint8_t *)pSrc;
    uint32_t num_chunks, frac_length;

    num_chunks = num_bytes >> 8;  // num_bytes / chunk_size
    frac_length = num_bytes - chunk_size * num_chunks;
    for(uint32_t n=0; n<num_chunks; n++){        
        bpi_serial_reset_buffers();                      // Empty buffers
        bpi_serial_write((uint8_t *)pSnd, chunk_size);   // Send data
        pSnd += chunk_size;
    }
    if(frac_length>0){
        bpi_serial_reset_buffers();                      // Empty buffers
        bpi_serial_write((uint8_t *)pSnd, frac_length);  // Send data
    }
}

/*
    Host wants to read an array of 4-byte data from the MCA
    after having send a command to the MCA.
    pDst: Pointer to a 4-byte aligned data source
    num_words: Number of 4-bye data to be written; data may be uint32_t or float.
    
    Data will be transmitted byte-wise with the LSB being sent first.
    At this stage it is assumed that the MCA is ready to send the requested data to the host.
    
*/
void bpi_read_buffer(uint32_t *pDst, uint32_t num_bytes){
    const uint32_t chunk_size = 256;
    uint8_t *pRcv = (uint8_t *)pDst;
    uint32_t num_chunks, frac_length;
    
    num_chunks = num_bytes >> 8;  // num_bytes / chunk_size
    frac_length = num_bytes - chunk_size * num_chunks;
    for(uint32_t n=0; n<num_chunks; n++){
        bpi_serial_read((uint8_t *)pRcv, chunk_size);  // Receive the data
        pRcv += chunk_size;
    }
    if(frac_length>0){
        bpi_serial_read((uint8_t *)pRcv, frac_length); // Receive the data
    }
    bpi_serial_reset_buffers();  // Empty buffers
}



// https://en.cppreference.com/w/c/language
// http://www.cplusplus.com/reference/clibrary/
