import json
import time
import zmq
import sys
sys.path.append("../../bpi_imports/")  # Make import search the 'bpi_import' directory also.
import bpi_communication


def log_count_rate(par):
    """
    We assume that only one device is attached, and that the MCA Data Server is running.

    Start a log of count rates to measure the response to quickly passing radioactive material.

    The purpose of the logger is to record quick changes of operating conditions,
    without polling the MCA at a high rate.

    A host that does not run a RTOS can not reliably poll the MCA-2000 at precise time intervals.
    Instead the logger captures up to 1023 measurements of two variables without requiring host action.

    The host can then read the log after it has been completed.  The logger implements a rolling log,
    meaning that old data will continuously be overwritten by new data.

    :param file_name: Write the voltage data to this file.
    :return:
    """
    time_step = 0.050  # Built-in time granularity is 50ms
    dwell_time = par["dwell_time"]  # in 50ms units; 0<= dwell_time <= 255
    wait_bins = par["wait_bins"]  # Wait that many logger dwell times before reading the logger data.
    index_0 = par["index_0"]  # Total count rate
    index_1 = par["index_1"]  # Count rate within the region of interest
    file_name = par["file_name"]
    # index_events = par["index_events"] # The MSB=1 indicates this is computed by the logger; It is not amongst the arm_status fields

    mds_client, sn = bpi_communication.connect_one_device()

    xctrl_0 = (index_1 << 16) + (index_0 << 8) + dwell_time  # DC-baseline, operating voltage, dwell_time
    cmd = {"type": "mca_cmd", "name": "arm_ctrl", "dir": "rmw", "sn": sn,
           "data": {"fields": {"xctrl_0": xctrl_0, 'run_action': 8}}}
    ret = mds_client.sar(cmd)

    time.sleep(wait_bins*dwell_time*time_step)

    cmd = {"type": "mca_cmd", "name": "arm_logger", "dir": "read", "sn": sn, 'num_items': 2048}  # For some versions of the PMT-1000, only 1024 items are available, in 2 511-item banks
    ret = mds_client.sar(cmd)

    var_0 = ret[sn]["user"]["var_0"]
    var_1 = ret[sn]["user"]["var_1"]
    dt = ret[sn]["fields"]["dt"]
    ch_0 = ret[sn]["fields"]["ch_0"]
    ch_1 = ret[sn]["fields"]["ch_1"]
    label_0 = par["labels"][ch_0]
    label_1 = par["labels"][ch_1]
    out_dict = {"dt": dt, "ch_0": ch_0, "ch_1": ch_1, "label_0": label_0, "label_1": label_1, "var_0": var_0, "var_1": var_1 }

    with open(file_name, 'w') as fout:
        fout.write(json.dumps(out_dict) + "\n")

    # Turn logging off
    xctrl_0 = 0  # Set at least the lower 8 bit to zero to turn logging off
    cmd = {"type": "mca_cmd", "name": "arm_ctrl", "dir": "rmw", "sn": sn,
           "data": {"fields": {"xctrl_0": xctrl_0, 'run_action': 8}}}
    ret = mds_client.sar(cmd)

labels= {
    0: "Operating Voltage",
    1: "Voltage_target",
    2: "Set Voltage",
    3: "Digital Gain Target",
    4: "CPU Temperature",
    5: "Sensor Temperature",
    6: "Average Temperature",
    7: "Wall Clock Time",
    8: "Run Status",
    9: "Sample Run Time",
    10: "Sample Count Rate",
    11: "Sample Count Rate Error",
    12: "Background Run Time",
    13: "Background Count Rate",
    14: "Background Count Rate Error",
    15: "Count Rate Difference",
    16: "Count Rate Difference Error",
    17: "Background Probability",
    18: "Background Alarmist Probability",
    19: "Background Cautious Probability",
    20: "Portal Ready",
    21: "Portal Alarm",
    22: "Portal Net Counts",
    23: "Portal Background Counts",
    24: "Portal Background Probability",
    25: "Portal Reset",
    26: "Region of Interest Count Rate",
    27: "Region of Interest Count Rate Error",
    28: "Region of Interest Count Rate",
    29: "Region of Interest Count Rate Error",
    30: "Region of Interest Count Rate Difference",
    31: "Region of Interest Count Rate Difference Error",
    32: "Region of Interest Sample Counts",
    33: "Region of Interest Background Counts",
    34: "Average LED or ROI Value",
    35: "DC-Baseline",
    36: "Trigger Level",
    128: "Logger Events",
    129: "Logger ROI Events",
    130: "Portal Net Counts",
    131: "Portal Signal Strength",
    132: "Sample Signal Strength"
}

par = {
    "dwell_time":   1,  # in 50ms units; 0 < dwell_time <= 255
    "wait_bins": 1024,  # Wait that many logger dwell times before reading the logger data.
    "index_0": 128,  # Total number of events within one logger time bin  (Computed, not part of arm_status fields)
    "index_1": 129,  # Number of events within ROI and one logger time bin  (Computed, not part of arm_status fields)
    "file_name": "../data/count_rate_log.json",
    "labels": labels
}
log_count_rate(par)

