import array
import json
import time
import shutil
import sys
sys.path.append("../../bpi_imports/")  # Make import search the 'bpi_import' directory also.
import bpi_communication as com

ARM_APP_ERASE = 1
ARM_APP_BOOT = 2

"""
    The ARM-based MCA-1000, MCA-2000, and MCA-3000 all boot their application using a boot loader.
    As shipped, the operation is invisible to the user.  However the user can update the
    application program by first invalidating the application program by sending a command
    to the application. On the next boot, the bootloader will recognize this and wait for
    a new application to be downloaded via USB.

    The code below is an example for managing the bootloader.  This code only communicates
    with the boot loader.  It will not communicate with the MCA application.

"""

def erase_arm():
    """
    We assume that only one MCA-1000 is attached, and that the MCA Data Server is running.

    When the boot loader is waiting for new code,
    use this function to erase the application program memory.

    :param arm_bin_file: Data file containing a valid FPGA configuration
    :return: None
    """
    mds_client, sn = connect_to_device()

    cmd_erase = {"type": "mca_cmd", "name": "arm_cmd", "dir": "short_write", "sn": sn, "ctrl": [ARM_APP_ERASE]}
    ret = mds_client.sar(cmd_erase)

    cmd_check = {"type": "mca_cmd", "name": "arm_status", "dir": "read", "sn": sn}
    # Wait until erase has completed.
    while True:
        time.sleep(0.010)
        ret = mds_client.sar(cmd_check)
        print(ret[sn]["registers"][0:5])
        if ret[sn]["fields"]["rom_busy"] == 0:
            break


def boot_arm():
    """
    We assume that only one MCA-1000 is attached, and that the MCA Data Server is running.

    After the boot loader has received new code,
    use this function to boot the application.

    :return: None
    """
    mds_client, sn = connect_to_device()

    cmd_boot = {"type": "mca_cmd", "name": "arm_cmd", "dir": "short_write", "sn": sn, "ctrl": [ARM_APP_BOOT]}
    ret = mds_client.sar(cmd_boot)


def write_arm(arm_bin_file):
    """
    We assume that only one MCA-1000 is attached, and that the MCA Data Server is running.

    When the boot loader is waiting for new code,
    use this function to download code the program memory.

    The code is encrypted, and can not be read back.

    :param arm_bin_file: Data file containing a valid FPGA configuration
    :return: None
    """
    print("Beginning app download")
    mds_client, sn = connect_to_device()

    # Read ARM code
    with open(arm_bin_file, "rb") as fin:
        arm_bin = fin.read()
    PL = 256  # page length of the memory
    FL = len(arm_bin) % PL  # fragment_length
    TL = ((len(arm_bin)-1) // PL + 1)*PL
    bytes_out = array.array('B', arm_bin)
    if FL > 0:
        bytes_out.extend([0]*(PL-FL))

    with open("./data/test.bin", "wb") as fout:
        fout.write(bytes_out)
    print("Length of config file = ", len(arm_bin))

    # Now, write the ARM code.

    cmd_to_buffer = {"type": "mca_cmd", "name": "arm_rom", "dir": "write", "sn": sn}
    cmd_check = {"type": "mca_cmd", "name": "arm_status", "dir": "read", "sn": sn}

    num_pages = len(bytes_out)//PL
    for n in range(num_pages):
        registers = bytes_out[n*PL: (n+1)*PL]

        cmd_to_buffer["data"] = {"registers": list(registers)}
        ret = mds_client.sar(cmd_to_buffer)
        time.sleep(0.020)

        # Make sure page has been written
        while True:
            time.sleep(0.010)
            ret = mds_client.sar(cmd_check)
            if ret[sn]["fields"]["rom_busy"] == 0:
                break


def read_status():
    """
        Read boot_ctrl
    """
    mds_client, sn = connect_to_device()
    cmd_read = {"type": "mca_cmd", "name": "arm_version", "dir": "read", "sn": sn}

    ret = mds_client.sar(cmd_read)
    #print(ret[sn])
    return ret[sn]


def connect_to_device():
    """  Collect here the repetitive steps of connecting to the MCA Data Server.
    """
    mds_ip = "tcp://127.0.0.1:5507"
    mds_client = com.zmq_device(mds_ip, "client")

    cmd = {"type": "server_cmd", "name": "hello"}
    ret = mds_client.sar(cmd)
    sn_list = ret['sn_list']
    for sn in sn_list:
        print("Serial number:", sn)
    sn = sn_list[0]  # We only use one attached device here
    return mds_client, sn


def load_and_boot(crypto_file):
    then = time.time()

    read_status()
    erase_arm()
    read_status()
    write_arm(crypto_file)
    time.sleep(2)
    read_status()
    boot_arm()

    duration = time.time()-then
    print("Programming time {:.0f} sec.".format(duration))

}

crypto_files = {
    "pmt1k": "../data/pmt2k_crypto_b19.bin",
    "sipm1k": "../data/sipm2k_crypto_b19.bin",
}
load_and_boot(crypto_files["pmt1k"])





