#!/usr/bin/python
#
from __future__ import division
import time
import math
import json
import time
import sys
sys.path.append("../../bpi_imports/")  # Make import search the 'bpi_import' directory also.
import bpi_communication
import histogram_analysis

version = 1.0


def log_histograms(par):
    """
    We assume that only one MCA-2000 is attached, and that the MCA Data Server is running.
    Note that 'sar' stands for send and receive.

    The function is used to acquire many histograms in a row.
    Set "save_histograms" to 1 to store the raw histograms in a potentially very big file.

    Set "live_analyis" to 1 to perform a peak fit and store temperature and peak positions,
    one line at a time, in a small file.

    """
    # Unpack the input parameters for use below
    histo_4k = par["histo_4k"] & 1
    dwell_time = par["dwell_time"]
    num_items = 2048*(1+histo_4k)  # 4096 vs 2048 MCA bins to read
    save_histograms = par["save_histograms"]  # 0-> Do not save the histograms; else do save the histograms
    num_histograms = par["num_histograms"]  # Number of histograms to acquire
    live_analysis = par["live_analysis"]  # 0-> no analysis; 1->Analyse each histogram as it was acquired
    fit_range = par["fit_range"]  # Function will perform a peak fit to the tallest peak in the range.
    fa = fit_range[0]
    fb = fit_range[1]

    # Open communicaion with MCA Data Server
    # 127.0.0.1 is the same as localhost
    mds_client = bpi_communication.zmq_device("tcp://127.0.0.1:5507", "client")  # Communicate with MDS

    # Get serial number
    cmd = {"type": "server_cmd", "name": "hello"}
    ret = mds_client.sar(cmd)
    sn_list = ret["sn_list"]
    for sn in sn_list:
        print("Unique serial number:", sn)
    sn = sn_list[0]  # We just serve one device here

    # List all commands
    cmd_start_mca = {"type": "mca_cmd", "name": "arm_ctrl",  "dir": "rmw", "sn": sn,
           "data": {"fields": {"run_time_sample": dwell_time},
           "user":{"histogram_run": 1, "active_bank": 0, "histo_4k": histo_4k,
                   "clear_statistics": 1, "clear_histogram": 1}}}
    cmd_histogram = {"type": "mca_cmd", "name": "arm_histogram", "dir": "read", "sn": sn, "num_items": num_items}
    cmd_arm_status = {"type": "mca_cmd", "name": "arm_status", "dir": "read", "sn": sn}
    cmd_arm_ctrl = {"type": "mca_cmd", "name": "arm_ctrl", "dir": "read", "sn": sn}

    for nhisto in range(num_histograms):
        # Start new sample histogram acquisition
        mds_client.sar(cmd_start_mca)
        time.sleep(dwell_time)

        # The histogram acquisition has completed; now read the data.

        # Read histogram data
        ret_histogram = mds_client.sar(cmd_histogram)
        histogram = ret_histogram[sn]["registers"]

        # Read arm_status, includes count rate data
        ret_status = mds_client.sar(cmd_arm_status)
        arm_status = ret_status[sn]

        # Read arm_ctrl, includes all instrument settings;
        ret_ctrl = mds_client.sar(cmd_arm_ctrl)
        arm_ctrl = ret_ctrl[sn]

        # We have all the data
        # Perform analysis
        if live_analysis:
            # Perform a peak fit
            fit_histogram, left_idx, fwhm = histogram_analysis.estimate_peak(histogram[fa:fb])
            fd = histogram_analysis.do_gauss_fit(fit_histogram, bck_model=2, fwhm=fwhm)
            peak_pos = fa + left_idx + fd["x_max"]
            fit_fwhm = fd["fwhm"]

            # Add data to the peaks output dictionary
            temperature = arm_status["fields"]["avg_temperature"]
            out_dict = {"temperature": temperature, "peak_pos": peak_pos, "fwhm": 100*fit_fwhm/peak_pos}

            # Append to the peaks log file
            with open(par["peaks_file_name"], "a") as fout:
                fout.write(json.dumps(out_dict)+"\n")


        # Now save histograms to file, if required.

        if save_histograms:
            # Write a json string, with all data; one line per histogram acquisition
            out_dict = {"sn":sn, "arm_ctrl": arm_ctrl, "arm_status": arm_status, "histogram": histogram}
            with open(par["histograms_file_name"], 'a') as fout:
                fout.write(json.dumps(out_dict) + "\n")


par = {  # Parameters controlling the log_histograms function
    "histo_4k": 0, # 0-> 2Kx32 histogram, 1-> 4Kx32 histogram
    "dwell_time": 10,  #  Data acquisition time in seconds
    "histograms_file_name": "../data/histograms.dat",  # Where to store the raw histograms
    "peaks_file_name": "../data/peaks.dat",  # Where to store the peak positions
    "save_histograms": 0,
    "num_histograms": 10,
    "live_analysis": 1,
    "fit_range": [300, 900]  # to safely catch the Cs-137 662keV peak, even if it shifts.

}
log_histograms(par)

