import time
import json
import sys
sys.path.append("../../bpi_imports/")  # Make import search the 'bpi_import' directory also.
import bpi_communication as com

"""
    This code is used to activate the bootloader by invalidating the application running on the MCA's processor.
    You need an encrypted ARM code file from Bridgeport Instruments to load a new application onto the MCA.

    When this file is executed the application is marked as invalid.  After the next power cycle, the bootloader
    will keep the MCA in bootloader mode, waiting for a new ARM application to be loaded via USB.

    This code only communicates with the application.  It can not communicate with the bootloader.

"""

class MCA_APP:

    def __init__(self):
        self.APP_CODE_OK = 0
        self.APP_CODE_INVALID = 1
        self.mds_ip = "tcp://127.0.0.1:5507"
        self.mds_client = None
        self.sn = None
        self.connect_to_device()

    def connect_to_device(self):
        """  Collect here the repetitive steps of connecting to the MCA Data Server.
        """
        self.mds_client = com.zmq_device(self.mds_ip, "client")
        cmd = {"type": "server_cmd", "name": "hello"}
        ret = self.mds_client.sar(cmd)
        sn_list = ret['sn_list']
        for sn in sn_list:
            print("Serial number:", sn)
        self.sn = sn_list[0]  # We only use one attached device here


    def get_version(self):
        """
            When the user application is launched, we can send an 'invalidate' command
            to force the boot loader to wait for a new app after the next power cycle
            or reset.  If the WDT is being used, it will force a software reset.
        """
        cmd = {"type": "mca_cmd", "name": "arm_version", "dir": "read", "sn": self.sn}
        ret = self.mds_client.sar(cmd)
        print(ret[self.sn]["registers"])
        return ret


    def is_valid(self):
        cmd = {"type": "mca_cmd", "name": "arm_version", "dir": "read", "sn": self.sn}
        regs = self.mds_client.sar(cmd)[self.sn]["registers"]
        ret = regs[15]
        #print(regs)
        ret = (ret>>24) & 0xFF
        return ret != 0xFF


    def invalidate(self):
        """
            When the user application is launched, we can send an 'invalidate' command
            to force the boot loader to wait for a new app after the next power cycle
            or reset.  If the WDT is being used, we could force a software reset.
        """
        cmd = {"type": "mca_cmd", "name": "arm_cmd", "dir": "short_write", "sn": self.sn, "ctrl": [self.APP_CODE_INVALID]}
        self.mds_client.sar(cmd)
        time.sleep(0.1)


    def validate(self):
        """
            When the user application is launched, we can send a 'validate' command
            to force the boot loader to accept the code and launch it again on next
            processor reset.
        """
        cmd = {"type": "mca_cmd", "name": "arm_cmd", "dir": "short_write", "sn": self.sn, "ctrl": [self.APP_CODE_OK]}
        self.mds_client.sar(cmd)
        time.sleep(0.1)


def validate_app():
    App = MCA_APP()
    if App.is_valid():
        print("Application is valid." )
    else:
        print("Application is invalid." )
    App.validate()
    if App.is_valid():
        print("Application is valid." )
    else:
        print("Application is invalid." )


def invalidate_app():
    App = MCA_APP()
    if App.is_valid():
        print("Application is valid." )
    else:
        print("Application is invalid." )
    App.invalidate()
    if App.is_valid():
        print("Application is valid." )
    else:
        print("Application is invalid." )

invalidate_app()
print(\
""">> 
Now unplug the MCA and plug it back in.
Wait for the OS to enumerate the USB bus.
Exit the MCA Data Sever and launch it again.""")



